# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import functools
import re
from typing import Any
from typing import TypeVar

from headless.ext.oauth2 import OIDCToken
from headless.ext.oauth2.models import SubjectIdentifier

from cbra.core.iam.models import Subject
from cbra.types import IDependant
from .models import AuthorizationServerModel
from .models import AuthorizationRequest
from .models import AuthorizationRequestParameters
from .models import AuthorizationState
from .models import ExternalAuthorizationState
from .models import Client
from .models import ManagedGrant
from .models import RefreshToken
from .models import ResourceOwner
from .types import AuthorizationCode
from .types import AuthorizationRequestIdentifier
from .types import IssuedAccessToken
from .types import PairwiseIdentifier


T = TypeVar('T', bound=Any)


class BaseStorage(IDependant):
    __module__: str = 'cbra.ext.oauth2'
    pattern: re.Pattern[str] = re.compile(r'((?<=[a-z0-9])[A-Z]|(?!^)[A-Z](?=[a-z]))')

    async def destroy(self, obj: AuthorizationServerModel) -> None:
        raise NotImplementedError

    async def get_oidc_token(self, oid: SubjectIdentifier) -> OIDCToken | None:
        raise NotImplementedError

    # TODO: Refactor everything below this line.
    async def get(
        self,
        cls: type[T],
        *args: Any,
        **kwargs: Any
    ) -> T | None:
        name = self.pattern.sub(r'_\1', cls.__name__)
        fn = f"get_{str.lower(name)}"
        if not hasattr(self, fn):
            raise NotImplementedError(
                f'{type(self).__name__}.{fn}() is not implemented.'
            )
        return await getattr(self, fn)(*args, **kwargs)

    async def get_authorization_request(
        self,
        oid: AuthorizationCode | AuthorizationRequestIdentifier
    ) -> AuthorizationRequest | None:
        params = await self.get(AuthorizationRequestParameters, oid)
        if params is None:
            return None
        return AuthorizationRequest.parse_obj(params)

    @functools.singledispatchmethod
    async def get_authorization_request_parameters(
        self,
        oid: AuthorizationCode | str
    ) -> AuthorizationRequestParameters | None:
        raise NotImplementedError(repr(oid))
    
    @get_authorization_request_parameters.register
    async def _get_authorization_request_by_code(
        self,
        oid: AuthorizationCode
    ) -> AuthorizationRequestParameters | None:
        return await self.get_authorization_request_by_code(oid)
    
    @get_authorization_request_parameters.register
    async def _get_authorization_request_by_id(
        self,
        oid: AuthorizationRequestIdentifier
    ) -> AuthorizationRequestParameters | None:
        return await self.get_authorization_request_by_id(oid)

    async def get_authorization_request_by_code(
        self,
        oid: AuthorizationCode
    ) -> AuthorizationRequestParameters | None: 
        raise NotImplementedError

    async def get_authorization_request_by_id(
        self,
        oid: AuthorizationRequestIdentifier
    ) -> AuthorizationRequestParameters | None: 
        raise NotImplementedError

    async def get_state(self, key: str) -> AuthorizationState | None:
        raise NotImplementedError

    @functools.singledispatchmethod
    async def persist(
        self,
        obj: AuthorizationServerModel | Subject
    ) -> None:
        raise TypeError(type(obj).__name__)

    async def persist_authorization_request(
        self,
        obj: AuthorizationRequestParameters
    ) -> None:
        raise NotImplementedError

    async def persist_client(self, obj: Client) -> None:
        raise NotImplementedError

    async def persist_issued_access_token(
        self,
        obj: IssuedAccessToken
    ) -> None:
        raise NotImplementedError

    async def persist_managed_grant(
        self,
        obj: ManagedGrant
    ) -> None:
        raise NotImplementedError

    async def persist_oidc_token(self, obj: OIDCToken) -> None:
        raise NotImplementedError

    async def persist_ppid(self, obj: PairwiseIdentifier) -> None:
        raise NotImplementedError

    async def persist_refresh_token(
        self,
        obj: RefreshToken
    ) -> None:
            raise NotImplementedError

    async def persist_resource_owner(self, obj: ResourceOwner) -> None:
        raise NotImplementedError

    async def persist_state(self, obj: AuthorizationState | ExternalAuthorizationState) -> None:
        raise NotImplementedError
    
    async def persist_subject(self, obj: Subject) -> None:
        raise NotImplementedError

    @persist.register
    async def _persist_authorization_request(
        self,
        obj: AuthorizationRequestParameters
    ) -> None:
        return await self.persist_authorization_request(obj)

    @persist.register
    async def _persist_client(
        self,
        obj: Client
    ) -> None:
        return await self.persist_client(obj)

    @persist.register
    async def _persist_external_state(
        self,
        obj: ExternalAuthorizationState
    ) -> None:
        return await self.persist_state(obj)

    @persist.register
    async def _persist_issued_access_token(
        self,
        obj: IssuedAccessToken
    ) -> None:
        return await self.persist_issued_access_token(obj)

    @persist.register
    async def _persist_managed_grant(
        self,
        obj: ManagedGrant
    ) -> None:
        return await self.persist_managed_grant(obj)

    @persist.register
    async def _persist_oidc_token(self, obj: OIDCToken) -> None:
        return await self.persist_oidc_token(obj)

    @persist.register
    async def _persist_ppid(
        self,
        obj: PairwiseIdentifier
    ) -> None:
        return await self.persist_ppid(obj)

    @persist.register
    async def _persist_refresh_token(
        self,
        obj: RefreshToken
    ) -> None:
        return await self.persist_refresh_token(obj)

    @persist.register
    async def _persist_resource_owner(
        self,
        obj: ResourceOwner
    ) -> None:
        return await self.persist_resource_owner(obj)

    @persist.register
    async def _persist_state(
        self,
        obj: AuthorizationState
    ) -> None:
        return await self.persist_state(obj)

    @persist.register
    async def _persist_subject(
        self,
        obj: Subject
    ) -> None:
        return await self.persist_subject(obj)


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import secrets
from datetime import datetime
from datetime import timezone

import pydantic

from ...types import AuthorizationRequestIdentifier
from ...types import ClientIdentifier


class BaseAuthorizationState(pydantic.BaseModel):
    client_id: ClientIdentifier
    created: datetime = pydantic.Field(
        default_factory=lambda: datetime.now(timezone.utc)
    )
    iss: str
    kind: str
    nonce: str = pydantic.Field(
        default_factory=lambda: secrets.token_hex(48)
    )
    redirect_uri: str
    return_url: str | None = None
    request_id: AuthorizationRequestIdentifier
    state: str = pydantic.Field(
        default_factory=lambda: secrets.token_hex(48)
    )

    def get_prompt_url(self) -> str:
        raise NotImplementedError


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import base64
from typing import Any

import fastapi
import pydantic
from headless.ext.shopify import v2023_1

from cbra.types import IVerifier
from cbra.types import Request
from cbra.ext.webhooks import WebhookEnvelope


DEFAULT_API_VERSION: str = '2023-01'
MODELS: dict[tuple[str, str], type[pydantic.BaseModel]] = {
    ('2023-01', 'orders'): v2023_1.Order
}


class ShopifyWebhookEnvelope(WebhookEnvelope):
    __module__: str = 'cbra.ext.shopify'
    api_version: str
    content: dict[str, Any]
    domain: str
    hmac: bytes | None = None
    event_name: str
    webhook_id: str | None = None

    def __init__(
        self,
        api_version: str | None = fastapi.Header(
            default=None,
            alias='X-Shopify-API-Version',
        ),
        domain: str = fastapi.Header(
            default=None,
            alias='X-Shopify-Shop-Domain',
        ),
        signature: str = fastapi.Header(
            default=None,
            alias='X-Shopify-Hmac-Sha256',
        ),
        topic: str = fastapi.Header(
            default=None,
            alias='X-Shopify-Topic'
        ),
        webhook_id: str | None = fastapi.Header(
            default=None,
            alias='X-Shopify-Webhook-Id'
        ),
        content: dict[str, Any] = fastapi.Body()
    ):
        self.api_version = api_version or DEFAULT_API_VERSION
        self.content = content
        self.domain = domain
        self.hmac = signature.encode('utf-8')
        self.event_name = topic
        self.resource, self.event = str.split(topic, '/')
        self.webhook_id = webhook_id

    def get_message(self) -> pydantic.BaseModel:
        return MODELS[(self.api_version, self.resource)].parse_obj(self.content)

    async def verify(self, request: Request, verifier: IVerifier) -> bool:
        if self.hmac is None:
            return False
        return await verifier.verify(
            signature=self.hmac,
            message=await request.body(),
            encoder=base64.b64encode
        )


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from .fatalclientexception import FatalClientException


class InvalidRequest(FatalClientException):
    __module__: str = 'cbra.ext.oauth2.types'

    def __init__(self, message: str | None = None):
        super().__init__(
            code='invalid_request',
            message=message or (
                'The request is missing a required parameter '
                'includes an invalid parameter value, includes a parameter more '
                'than once, or is otherwise malformed.'
            )
        )



# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Union

import pydantic

from .emailscope import EmailScope
from .genericscope import GenericScope
from .openidscope import OIDCScope
from .profilescope import ProfileScope
from .subscope import SubScope
from .useragentfingerprintscope import UserAgentFingerprint


class GrantedScope(pydantic.BaseModel):
    __root__: Union[
        EmailScope,
        OIDCScope,
        ProfileScope,
        SubScope,
        UserAgentFingerprint,
        GenericScope # IMPORTANT: must be last
    ]

    @property
    def name(self) -> str:
        return self.__root__.name


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import secrets
from typing import Any

from cbra.types import Request
from headless.ext.oauth2 import Client


class DanceInitiationMixin:
    __module__: str = 'cbra.ext.oauth2'
    client: Client
    cookie_prefix: str
    redirection_endpoint: str
    request: Request

    async def create_authorization_request(
        self,
        scope: set[str],
        state: str | None = None,
        redirect_uri: str | None = None,
        resources: set[str] | None = None,
        **params: Any
    ) -> tuple[str, str, str]:
        state = state or secrets.token_urlsafe(48)
        nonce = secrets.token_urlsafe(48)
        return await self.client.authorize(
            redirect_uri=redirect_uri or self.get_redirect_uri(),
            state=state,
            nonce=nonce,
            scope=scope,
            resources=resources,
            **params
        ), state, nonce
    
    def get_redirect_uri(self) -> str:
        return str(self.request.url_for(self.redirection_endpoint))


# Copyright (C) 2022 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import os
from typing import Any
os.environ['PYTHON_SETTINGS_MODULE'] = __name__

import uvicorn

from cbra.ext import oauth2
from cbra.ext.google import Service


APP_ISSUER: str = 'https://localhost.webiam.id'

APP_CLIENT_ID: str = 'self'

APP_CLIENT_SECRET = SECRET_KEY = 'changeme'

DEBUG: bool = True

OAUTH2_CLIENTS: list[dict[str, Any]] = [
    {
        'name': "google",
        'issuer': 'https://accounts.google.com',
        'allowed_email_domains': ['gmail.com'],
        'credential': {
            'kind': 'ClientSecret',
            'client_id': os.environ['WEBID_GOOGLE_CLIENT_ID'],
            'client_secret': os.environ['WEBID_GOOGLE_CLIENT_SECRET']
        }
    },
    {
        'name': "microsoft",
        'issuer': 'https://login.microsoftonline.com/consumers/v2.0',
        'allowed_email_domains': ['hotmail.com', 'live.com', 'outlook.com'],
        'credential': {
            'kind': 'ClientSecret',
            'client_id': os.environ['WEBID_MICROSOFT_CLIENT_ID'],
            'client_secret': os.environ['WEBID_MICROSOFT_CLIENT_SECRET']
        }
    },
    {
        'name': "yahoo",
        'issuer': 'https://api.login.yahoo.com',
        'allowed_email_domains': ['yahoo.com'],
        'credential': {
            'kind': 'ClientSecret',
            'client_id': os.environ['WEBID_YAHOO_CLIENT_ID'],
            'client_secret': os.environ['WEBID_YAHOO_CLIENT_SECRET']
        }
    },
]


app: Service = Service()
app.add(
    oauth2.AuthorizationServer(
        client=True,
        #downstream={
        #    'protocol': 'oidc',
        #    'allowed_email_domains': [
        #        'buymolano.com',
        #        'molano.nl',
        #        'refurbcity.nl'
        #    ],
        #    'name': 'molano',
        #    'contacts': [
        #        'Noek de Feber <n.defeber@molano.nl>',
        #        'Cochise Ruhulessin <c.ruhulessin@.molano.nl>',
        #    ],
        #    'credential': {
        #        'kind': 'ClientSecret',
        #        'client_id': os.environ['MOLANO_WEBIAM_CLIENT_ID'],
        #        'client_secret': os.environ['MOLANO_WEBIAM_CLIENT_SECRET'],
        #    },
        #    'issuer': 'https://accounts.google.com',
        #    'scope': {'email', 'openid', 'profile'},
        #    'params': {'hd': 'molano.nl'}
        #},
        response_types=[
            oauth2.types.ResponseType.code
        ]
    ),
    prefix='/oauth2'
)


@app.on_event('startup')
def f():
    print('https://localhost.webiam.id/oauth2/authorize?client_id=self&response_type=code&state=foo')


if __name__ == '__main__':
    uvicorn.run('__main__:app', reload=True, port=6004) # type: ignore


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import logging
from typing import Awaitable
from typing import Callable
from typing import TypeVar

import fastapi
from fastapi.security import HTTPAuthorizationCredentials
from fastapi.security.utils import get_authorization_scheme_param
from headless.ext import oauth2
from headless.ext.oauth2.models import ClaimSet
from cbra.types import IDependant
from cbra.types import IPrincipal
from cbra.types import ISubject
from cbra.types import Request
from cbra.types import ServiceNotAvailable
from ..types import BearerTokenException
from ..types import IssuedAccessTokenIdentifier
from ..types import RFC9068AccessToken


P = TypeVar('P', bound='RequestAccessToken')
logger: logging.Logger = logging.getLogger('cbra.endpoint')


class RequestAccessToken(IDependant, IPrincipal):
    __module__: str = 'cbra.ext.oauth2.params'
    authorization: HTTPAuthorizationCredentials | None
    supported_schemes: set[str] = {"bearer"}
    token: RFC9068AccessToken | None = None

    @property
    def token_id(self) -> IssuedAccessTokenIdentifier | None:
        if self.authorization is not None:
            return IssuedAccessTokenIdentifier.parse_token(self.authorization.credentials)

    def __init__(
        self,
        header: str | None = fastapi.Header(
            default=None,
            alias='Authorization',
            title="Authorization",
            description="An RFC 9068 access token."
        )
    ):
        self.authorization = None
        scheme, credentials = get_authorization_scheme_param(header)
        if not scheme and credentials:
            raise BearerTokenException(
                'Bearer', 'invalid_token', "Malformed Authorization header."
            )
        if scheme:
            if str.lower(scheme) not in self.supported_schemes:
                raise BearerTokenException('Bearer', 'invalid_token', 'Unsupported scheme')
            if not credentials:
                raise BearerTokenException('Bearer', 'invalid_token', 'Missing token')
            self.authorization = HTTPAuthorizationCredentials(
                scheme=scheme,
                credentials=credentials
            )
            self.token = RFC9068AccessToken.parse_jwt(self.authorization.credentials)

    def is_anonymous(self) -> bool:
        return self.authorization is None
    
    def get_client(self) -> oauth2.Client:
        assert self.token is not None
        return oauth2.Client(issuer=self.token.iss)

    def has_audience(self) -> bool:
        return False

    async def resolve(
        self,
        resolve: Callable[..., Awaitable[ISubject]]
    ) -> ISubject:
        if self.token is None:
            raise BearerTokenException(
                scheme='Bearer',
                error='invalid_request',
                error_description=(
                    "The request did not provide an RFC 9068 access token "
                    "or other credential that is understood by the server."
                ),
                status_code=400
            )
        return await resolve(self)

    async def userinfo(self) -> ClaimSet:
        assert self.authorization is not None
        assert self.token is not None
        async with self.get_client() as client:
            if not client.server.userinfo_endpoint:
                raise BearerTokenException(
                    scheme='Bearer',
                    error='invalid_token',
                    error_description=(
                        "The authorization server that issued the token does not expose "
                        "the OpenID Connect UserInfo endpoint."
                    )
                )
            try:
                return await client.userinfo(self.authorization.credentials)
            except Exception as exc:
                logger.critical(
                    "Caught fatal %s while requesting userinfo from %s",
                    type(exc).__name__, self.token.iss
                )
                raise ServiceNotAvailable

    async def verify(
        self,
        request: Request,
        trusted_issuers: set[str],
        scope: set[str] | None = None
    ):
        """Verifies that the token is accepted by the resource
        server. Raise :exc:`~cbra.ext.oauth2.types.BearerTokenException`
        if the token does not verify.
        """
        assert self.authorization is not None
        assert self.token is not None

        # Ensure that the audience, issuer and scope are accepted by the
        # resource server.
        self.token.verify(
            audiences={
                f'{request.url.scheme}://{request.url.netloc}',
                str(request.url)
            },
            issuers=trusted_issuers,
            scope=scope
        )

        # Verify the signature if all other aspects of the token are verified,
        # to prevent unnecessary remote calls.
        async with self.get_client() as client:
            if not await client.verify_signature(self.authorization.credentials):
                raise BearerTokenException(
                    scheme='Bearer',
                    error='invalid_token',
                    error_description=(
                        "The signature of the access token could not be verified "
                        "using the known keys of the authorization server specified "
                        "in the 'iss' claim."
                    )
                )


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import urllib.parse

from cbra.core.conf import settings
from cbra.types import Forbidden
from .danceinitiationmixin import DanceInitiationMixin
from .models import FrontendLoginRequest
from .models import FrontendLoginResponse
from .tokenhandlerendpoint import TokenHandlerEndpoint


class FrontendLoginEndpoint(TokenHandlerEndpoint, DanceInitiationMixin):
    __module__: str = 'cbra.ext.oauth2'
    name: str = 'bff.login'
    path: str = '/oauth/v2/login'
    redirection_endpoint: str = 'bff.redirection'
    status_code: int = 303
    summary: str = 'Frontend Login Endpoint'

    def can_redirect(self, redirect_uri: str) -> bool:
        """Return a boolean if the user-agent may request a redirect to the
        given URI.
        """
        p = urllib.parse.urlparse(redirect_uri)
        return any([
            str.startswith(redirect_uri, '/'),
            all([
                p.netloc == self.request.url.netloc,
                p.scheme == self.request.url.scheme,
            ]),
            p.netloc in settings.LOGIN_AUTHORIZED_DOMAINS
        ])
    
    def get_resources(self) -> set[str]:
        return {spec['resource'] for spec in settings.OAUTH2_RESOURCE_SERVERS.values()}

    async def get(self) -> FrontendLoginResponse:
        """Retrieve an OIDC ID Token from the trusted authorization
        server to establish the identity of the caller.
        """
        return await self.redirect_user_agent('/', {'email', 'openid', 'profile'})
    
    async def redirect_user_agent(self, redirect_uri:str, scope: set[str]) -> FrontendLoginResponse:
        if not self.can_redirect(redirect_uri):
            raise Forbidden
        scope = scope or {'email', 'openid', 'profile'}
        for resource in settings.OAUTH2_RESOURCE_SERVERS.values():
            scope.update(resource.get('scope') or set())
        url, state, nonce = await self.create_authorization_request(
            scope=scope,
            access_type='offline',
            resources=self.get_resources()
        )
        self.set_cookie('nonce', nonce, httponly=True, secure=True)
        self.set_cookie('scope', str.join(' ', sorted(scope)), httponly=True, secure=True)
        self.set_cookie('state', state, httponly=True, secure=True)
        self.set_cookie('redirect_uri', redirect_uri, httponly=True, secure=True)
        return FrontendLoginResponse(redirect_to=url)

    async def post(self, dto: FrontendLoginRequest) -> FrontendLoginResponse:
        """Retrieve an OIDC ID Token from the trusted authorization
        server to establish the identity of the caller.
        """
        return await self.redirect_user_agent(dto.redirect_uri, dto.scope)


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Any
from typing import Protocol
from typing import TypeVar


T = TypeVar('T', bound='IRefreshToken')


class IRefreshToken(Protocol):
    __module__: str = 'cbra.ext.oauth2.types'
    auth_time: int
    claims: dict[str, Any]
    ppid: int
    scope: set[str] | list[str]
    token: str
    sub: int

    def allows_resource(self, resource: str) -> bool: ...
    def allows_scope(self, scope: set[str]) -> bool: ...
    def is_active(self) -> bool: ...
    def refresh(self: T) -> T: ...


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Any
from typing import AsyncIterable

import fastapi
from ckms.core import Keychain
from headless.ext.oauth2 import Client
from headless.ext.oauth2.models import ClientAuthenticationMethod

from cbra.core.conf import settings
from cbra.core.params import ApplicationKeychain
from cbra.types import Request


__all__: list[str] = ['ApplicationClient']


def uses_client_secret(method: ClientAuthenticationMethod):
    return method in {
        ClientAuthenticationMethod.client_secret_basic,
        ClientAuthenticationMethod.client_secret_post
    }


def uses_client_jwt_assertion(method: ClientAuthenticationMethod) -> bool:
    return method == ClientAuthenticationMethod.private_key_jwt


async def get(
    request: Request,
    keychain: Keychain = ApplicationKeychain
) -> AsyncIterable[Client]:
    domain: str = request.url.netloc
    client_id: str | None = settings.APP_CLIENT_DOMAINS.get(domain) or settings.APP_CLIENT_ID
    if client_id is None:
        raise ValueError("Configure the APP_CLIENT_ID setting")
    method = ClientAuthenticationMethod(settings.APP_CLIENT_AUTHENTICATION_METHOD)
    params: dict[str, Any] = {
        'issuer': settings.APP_ISSUER,
        'client_auth': method,
        'client_id': client_id,
        'trust_email': settings.APP_ISSUER_TRUST
    }
    if uses_client_secret(method):
        params['client_secret'] = settings.APP_CLIENT_SECRET
    elif uses_client_jwt_assertion(method):
        params.update({
            'signing_key': keychain.get('sig'),
            'encryption_key': keychain.get('enc')
        })
    else:
        raise NotImplementedError
    async with Client(**params) as client:
        yield client


ApplicationClient: Client = fastapi.Depends(get)


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import os

os.environ['PYTHON_SETTINGS_MODULE'] = __name__

OAUTH2_CLIENTS = [
    {
        'name': 'google',
        'issuer': 'https://accounts.google.com',
        'client_id': '641438517445-22t22tlpnbk4a7575a89js6csko51e6a.apps.googleusercontent.com',
        'client_secret': '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',
        'trust_email': True,
        'scope': {'email', 'profile'}
    }
]

SECRET_KEY: str = 'changeme'

from typing import Any

import cbra.core as cbra
from cbra.ext import google
from cbra.ext.oauth2 import OIDCRegistrationEndpoint
from cbra.ext.oauth2 import LoginEndpoint


class HomeEndpoint(cbra.Endpoint):

    async def get(self) -> dict[str, Any]:
        await self.session
        return {
            'claims': self.session.claims.dict()
        }


app = google.Service(docs_url='/ui')
app.add(OIDCRegistrationEndpoint, path='/oauth2/callback/{client_id}')
app.add(LoginEndpoint, path='/oauth2/login/{client_id}')
app.add(HomeEndpoint, path='/')

OAUTH2_STATE: str = os.urandom(32).hex()


@app.on_event('startup')
async def boot():
    url = "http://localhost:8000/oauth2/login/google"
    print(f"Open {url} to initiate the flow.")


if __name__ == '__main__':
    import uvicorn
    uvicorn.run('__main__:app', reload=True) # type: ignore



# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import functools
import json
from typing import get_origin
from typing import Any
from typing import AsyncGenerator
from typing import TypeVar

from google.cloud.datastore import Client
from google.cloud.datastore import Entity
from google.cloud.datastore import Key
from google.cloud.datastore import Query

from cbra.core import ioc
from cbra.ext.security import ApplicationKeychain
from cbra.types import PersistedModel
from .datastorequeryresult import DatastoreQueryResult
from .runner import Runner


M = TypeVar('M', bound=PersistedModel)
Q = TypeVar('Q', bound=PersistedModel)
R = TypeVar('R')


class DatastoreRepository(Runner):
    __module__: str = 'cbra.ext.google'
    client: Client
    keychain: ApplicationKeychain
    model: type[Any]

    def __init__(
        self,
        client: Client | Any = ioc.inject('GoogleDatastoreClient'),
        keychain: ApplicationKeychain = ApplicationKeychain.depends()
    ):
        if not isinstance(client, Client):
            raise TypeError(f"Invalid client: {repr(client)}")
        self.client = client
        self.keychain = keychain

    async def allocate(self, model: type[Any], n: int = 1) -> list[int]:
        base = self.key(kind=model.__name__)
        return [
            x.id for x in
            await self.run_in_executor(
                functools.partial( # type: ignore
                    self.client.allocate_ids, # type: ignore
                    incomplete_key=base,
                    num_ids=n
                )
            )
        ]

    async def delete(self, keys: Key | list[Key]) -> None:
        func = self.client.delete if isinstance(keys, Key) else self.client.delete_multi # type: ignore
        return await self.run_in_executor(functools.partial(func, keys)) # type: ignore

    async def execute(self, query: Query, _R: type[R]) -> AsyncGenerator[R, None]:
        if _R == Key:
            query.keys_only()
        origin = get_origin(_R)
        for obj in await self.run_in_executor(query.fetch): # type: ignore
            if origin == int:
                yield obj.key.id
            elif origin in (Entity, dict):
                yield obj
            elif origin == Key:
                yield obj.key

    async def fetch(
        self,
        model: type[Q],
        query: Query,
        cursor: str | None = None,
        limit: int = 100
    ) -> DatastoreQueryResult[Q]:
        """Like :meth:`execute()`, but returns a :class:`QueryResult` instance
        holding the pagination token and the items matching the query.
        """
        result = await self.run_in_executor(
            functools.partial(query.fetch, start_cursor=cursor, limit=limit) # type: ignore
        )
        return DatastoreQueryResult(
            token=result.next_page_token, # type: ignore
            entities=[self.restore(model, x) for x in result]
        )

    def key(
        self,
        kind: str | Any,
        entity_id: int | str | None = None,
        parent: Key | None = None
    ) -> Key:
        if not isinstance(kind, str):
            kind = kind.__name__
        assert isinstance(kind, str), repr(kind)
        key = self.client.key(kind, parent=parent) # type: ignore
        if isinstance(entity_id, (int, str)):
            key = self.client.key(kind, entity_id, parent=parent) # type: ignore
        return key
    
    def model_key(self, instance: PersistedModel) -> Key:
        Model: type[PersistedModel] = type(instance)
        if instance.__surrogate__ != NotImplemented:
            key = self.key(Model.__name__, instance.__surrogate__)
        elif len(instance.__key__) == 1:
            key = self.key(Model.__name__, instance.__key__[0][1]) # type: ignore
        else:
            raise NotImplementedError
        return key

    async def get_entity_by_key(self, key: Key) -> Entity | None:
        return await self.run_in_executor(
            functools.partial(
                self.client.get, # type: ignore
                key=key
            )
        )

    async def get_model_by_key(
        self,
        model: type[M],
        key: int | str | Key
    ) -> M | None:
        if isinstance(key, (int, str)):
            key = self.key(model.__name__, key)
        assert isinstance(key, Key)
        entity = await self.get_entity_by_key(key)
        if entity is None:
            return None
        return self.restore(model, entity)

    async def one(self, model: type[Q], query: Query) -> Q | None:
        """Lookup exactly one object."""
        instance: Q | None = None
        for entity in await self.run_in_executor(query.fetch): # type: ignore
            if instance is not None:
                raise RuntimeError("Multiple objects returned.")
            instance = self.restore(model, entity)
        return instance

    async def put(self, instance: PersistedModel, exclude: set[str] | None = None) -> Key:
        Model: type[PersistedModel] = type(instance)
        exclude = set(exclude or set()) | {'__surrogate__'}
        field: str | None = None
        if instance.__surrogate__ != NotImplemented:
            field = Model.__surrogate__.attname # type: ignore

            assert field is not None
            if instance.__surrogate__ is None:
                setattr(instance, field, (await self.allocate(Model))[0])

            # Exclude the surrogate key since it is retained in the Google Datastore
            # entities' key.
            exclude.add(field)
        elif len(instance.__key__) == 1:
            exclude.add(instance.__key__[0][0]) # type: ignore
        else:
            raise NotImplementedError
        key = self.model_key(instance)
        data = json.loads(instance.json(exclude=exclude))
        assert instance.__surrogate__ is not None
        entity = Entity(key=key)
        entity.update(data) # type: ignore
        await self.run_in_executor(self.client.put, entity) # type: ignore
        return entity.key # type: ignore

    def construct_key(self, model: Any) -> Key:
        raise NotImplementedError

    def query(self, kind: str | type[Q], ancestor: Key | None = None) -> Query:
        if not isinstance(kind, str):
            kind = kind.__name__
        return self.client.query(kind=kind, ancestor=ancestor) # type: ignore

    def restore(self, model: type[Q], entity: dict[str, Any] | Entity) -> Q:
        if model.__surrogate__ != NotImplemented: # type: ignore
            k = model.__surrogate__.attname # type: ignore
            v = entity.key.id # type: ignore
        elif model.__key__ != NotImplemented and len(model.__key__) == 1: # type: ignore
            k = model.__key__[0][0] # type: ignore
            v = entity.key.id or entity.key.name # type: ignore
        else:
            raise NotImplementedError
        return model.parse_obj({
            **entity,
            k: v
        })


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import fastapi
from headless.ext.oauth2.models import ClaimSet

from cbra.types import NotFound
from cbra.types import SessionRequestPrincipal
from .endpoints import AuthorizationServerEndpoint
from .models import AuthorizationRequest
from .models import CurrentAuthorizationRequest
from .models import PatchAuthorizationRequest
from .models import PatchAuthorizationResponse
from .types import AuthorizationRequestIdentifier
from .types import RedirectURI


class AuthorizationRequestEndpoint(AuthorizationServerEndpoint):
    __module__: str = 'cbra.ext.oauth2'
    name: str = 'oauth2.authorize'
    principal: SessionRequestPrincipal # type: ignore
    path: str = '/requests'
    request_id: AuthorizationRequestIdentifier = fastapi.Cookie(
        default=...,
        title="Request ID",
        alias='oauth2.request',
        description=(
            "The authorization request identifier. This cookie is set by the "
            "authorization endpoint in the case that the resource owner "
            "needs to perform a certain action."
        )
    )
    summary: str = 'Authorization Request'

    async def get_authorization_request(self) -> AuthorizationRequest:
        await self.session
        request = await self.storage.get(AuthorizationRequest, self.request_id)
        if request is None:
            self.logger.debug("Authorization request does not exist")
            raise NotFound
        if request.session_id != self.session.id:
            self.logger.debug(
                "Authorization request lookup by unknown session (expected: %s, actual: %s)",
                request.session_id, self.session.id
            )
            raise NotFound
        return request

    async def get(self) -> CurrentAuthorizationRequest:
        request = await self.get_authorization_request()
        return CurrentAuthorizationRequest(
            client=request.client_info,
            consent=request.consent,
            email=request.email,
            scope=request.scope,
            id_token=ClaimSet.parse_obj({'sub': '0', **request.id_token})
        )
    
    async def patch(self, dto: PatchAuthorizationRequest) -> PatchAuthorizationResponse:
        request = await self.get_authorization_request()
        url = None
        if dto.deny:
            redirect_uri = RedirectURI(request.redirect_uri)
            url = redirect_uri.redirect(error='access_denied')
        else:
            raise NotImplementedError
        assert url is not None
        return PatchAuthorizationResponse(next=url)


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import fastapi

from cbra.core.conf import settings
from cbra.types import Request


__all__: list[str] = [
    'CurrentIssuer'
]


def get_issuer(request: Request) -> str:
    return settings.OAUTH2_ISSUER or\
        f'{request.url.scheme}://{request.url.netloc}'


CurrentIssuer: str = fastapi.Depends(get_issuer)


# Copyright (C) 2021-2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Any
from typing import Callable
from typing import Coroutine

import pytest

import cbra.core as cbra
from cbra.ext import webhooks
from .base import *
from .conftest import AuthenticatedBookWebhookEndpoint


@pytest.fixture # type: ignore
def app() -> cbra.Application:
    app = cbra.Application()
    app.add(AuthenticatedBookWebhookEndpoint, path='/')
    return app


@pytest.fixture
def sign() -> bool:
    return True


@pytest.mark.asyncio
async def test_send_message_unsigned(
    request_factory: Callable[..., Coroutine[Any, Any, webhooks.WebhookResponse]]
):
    result = await request_factory(
        topic='hello',
        msg={'foo': 1},
        url='/',
        sign=False
    )
    assert not result.success
    assert result.code == 'INVALID_SIGNATURE'


# Copyright (C) 2021-2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
"""Tests that a model field that is declared as optional is also allowed
on the response model.
"""
from typing import Any

import pytest
import pytest_asyncio
from headless.core import httpx

import cbra.core as cbra


class Endpoint(cbra.Endpoint):
    versioned: bool = True
    test: bool = True

    async def post(self, data: dict[str, Any]) -> None:
        assert set(data.get('If-Match') or []) == self.etag


@pytest_asyncio.fixture(scope='session') # type: ignore
async def client():
    app = cbra.Application()
    app.add(Endpoint, path='/')
    async with httpx.Client(base_url='https://cbra', app=app) as client:
        yield client


@pytest.mark.asyncio
async def test_etag_single_value(
    client: httpx.Client
):
    response = await client.post(
        url='/',
        headers={'If-Match': 'foo'},
        json={'If-Match': ['foo']}
    )
    assert 200 <= response.status_code < 300, response.content


@pytest.mark.asyncio
async def test_etag_multiple_value(
    client: httpx.Client
):
    response = await client.post(
        url='/',
        headers={'If-Match': 'foo,bar'},
        json={'If-Match': ['foo', 'bar']}
    )
    assert 200 <= response.status_code < 300, response.content


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Any

import pydantic


class AuthorizationRequestObject(pydantic.BaseModel):
    request: str
    remote_host: str

    async def load(
        self,
        client: Any,
        storage: Any,
        session_id: str
    ):
        raise NotImplementedError


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import fastapi

from cbra.types import Abortable
from .webhookresponse import WebhookResponse


class WebhookException(Abortable):
    __module__: str = 'cbra.ext.webhooks'
    status_code: int = 200
    code: str
    reason: str

    async def as_response(self) -> fastapi.responses.JSONResponse:
        return fastapi.responses.JSONResponse(
            status_code=self.status_code,
            content=WebhookResponse(
                code=self.code,
                reason=self.reason,
                accepted=False,
                success=False
            ).dict()
        )



# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import urllib.parse

import fastapi
from headless.ext import oauth2

import cbra.core as cbra
from cbra.core.conf import settings
from cbra.types import Forbidden
from cbra.types import NotFound
from .endpoints import AuthorizationServerEndpoint
from .types import LoginResponse
from .models import AuthorizationState


class LoginEndpoint(AuthorizationServerEndpoint):
    __module__: str = 'cbra.ext.oauth2'
    name: str = 'oauth2.login'
    nonce_cookie_name: str = 'oidc.nonce'
    path: str = '/login'
    redirect_cookie_name: str = 'oauth2.redirect-uri'
    state_cookie_name: str = 'oauth2.state'
    status_code: int = 303
    summary: str = 'Login Endpoint'
    tags: list[str] = ['OAuth 2.x/OpenID Connect']

    async def get(
        self,
        success_url: str | None = fastapi.Query(
            default=None,
            title='Next URL',
            alias='next',
            description=(
                'The URL to which the server must redirect the user-agent '
                'after a successful login. If this value is not provided '
                'then the server may use a default URL.'
            )
        ),
        client_id: str = fastapi.Path(
            default=...,
            title='Client ID',
            description=(
                'The OAuth 2.x/OpenID Connect client to initiate the '
                'authorization flow with. If not provided, the server '
                'may select a default client.'
            )
        )
    ) -> fastapi.responses.RedirectResponse:
        """Initiate the authorization code flow by redirecting the user agent
        to the authorization servers' authorization endpoint.
        """
        success_url = success_url or '/'
        if not self.can_redirect(success_url):
            raise Forbidden
        if self.ctx.is_authenticated():
            return fastapi.responses.RedirectResponse(
                status_code=303,
                url=success_url
            )

        client = await self.get_client(client_id)
        if client is None:
            raise NotFound
        params = AuthorizationState.new(redirect_uri=success_url)
        redirect_uri = self.request.url_for('oauth2.callback', client_id=client_id)
        async with client:
            url = await client.authorize(
                state=params.state,
                redirect_uri=str(redirect_uri),
                scope={'openid', 'email'},
                nonce=params.nonce,
                prompt='select_account'
            )

        # Persist the state so that the callback endpoint can find the needed
        # information to forward the request.
        await self.storage.persist(params)
        return fastapi.responses.RedirectResponse(
            status_code=self.status_code,
            url=url
        )

    @cbra.describe(status_code=200)
    async def post(
        self,
        client_id: str = fastapi.Path(
            default=...,
            title='Client ID',
            description=(
                'The OAuth 2.x/OpenID Connect client to initiate the '
                'authorization flow with. If not provided, the server '
                'may select a default client.'
            )
        )
    ) -> LoginResponse:
        """Initiate the authorization code flow and return a JSON object
        containing information on how to proceed.
        """
        client = await self.get_client(client_id=client_id)
        if client is None:
            raise NotFound
        raise NotImplementedError

    def can_redirect(self, redirect_uri: str) -> bool:
        """Return a boolean if the user-agent may request a redirect to the
        given URI.
        """
        p = urllib.parse.urlparse(redirect_uri)
        return any([
            str.startswith(redirect_uri, '/'),
            all([
                p.netloc == self.request.url.netloc,
                p.scheme == self.request.url.scheme,
            ]),
            p.netloc in settings.LOGIN_AUTHORIZED_DOMAINS
        ])

    async def get_client(self, client_id: str) -> oauth2.Client | None:
        """Return a preconfigured OAuth 2.x/OpenID Connect client,
        or ``None`` if the client does not exist.
        """
        # TODO: Quite ugly
        for client in settings.OAUTH2_CLIENTS:
            if client_id in {client.get('name'), client.get('client_id')}:
                instance = oauth2.Client(**client)
                break
        else:
            instance = None
        return instance


# Copyright (C) 2021-2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import hmac
from typing import Any

import pytest
import pytest_asyncio
from ckms.utils import b64encode_str
from headless.core import httpx

import cbra.core as cbra
from cbra.core.conf import settings
from cbra.types import Session


async def sign_hmac(value: bytes) -> str:
    return b64encode_str(hmac.digest(b'foo', value, 'sha384'))


def test_set_clears_hmac():
    session = Session.new()
    session.hmac = 'foo'
    session.set('sub', 'bar')
    assert session.hmac is None


def test_set_returns_modified():
    session = Session.new()
    assert session.set('sub', 'bar')
    assert not session.set('sub', 'bar')


# TODO: move to other file
@pytest.mark.asyncio
async def test_no_session_without_modify_sets_session(
    client: httpx.Client
):
    response = await client.get(url='/')
    assert settings.SESSION_COOKIE_NAME in response.impl.cookies


@pytest.mark.asyncio
async def test_session_updates_are_returned(
    client: httpx.Client
):
    response = await client.post(url='/')
    assert settings.SESSION_COOKIE_NAME in response.impl.cookies
    obj = Session.parse_cookie(response.impl.cookies[settings.SESSION_COOKIE_NAME])
    assert obj is not None
    assert obj.claims is not None
    assert obj.claims.iss == 'foo'
    assert obj.claims.sub == 'bar'


@pytest.mark.asyncio
async def test_session_invalid_key_is_rotated(
    client: httpx.Client
):
    response = await client.post(url='/')
    assert settings.SESSION_COOKIE_NAME in response.impl.cookies
    obj = Session.parse_cookie(response.impl.cookies[settings.SESSION_COOKIE_NAME])
    assert obj is not None
    print(f"Estblished session with HMAC {obj.hmac}")

    sig = obj.hmac
    evil = Session.parse_cookie(response.impl.cookies[settings.SESSION_COOKIE_NAME])
    assert evil is not None
    evil.set('sub', 'baz')
    await evil.sign(sign_hmac)
    assert evil.hmac != sig
    assert evil.as_cookie() != obj.as_cookie()
    print(f"Attacker created session with HMAC {evil.hmac}")

    # The original session must validate
    response = await client.get(
        url='/',
        cookies={settings.SESSION_COOKIE_NAME: response.impl.cookies[settings.SESSION_COOKIE_NAME]}
    )
    assert settings.SESSION_COOKIE_NAME not in response.impl.cookies

    # The tampered session rotates to a new session
    print("Sending evil session")
    client.cookies.clear()
    response = await client.get(
        url='/',
        cookies={settings.SESSION_COOKIE_NAME: evil.as_cookie()}
    )
    assert settings.SESSION_COOKIE_NAME in response.impl.cookies
    session = Session.parse_cookie(response.impl.cookies[settings.SESSION_COOKIE_NAME])
    assert session is not None
    assert session.id != evil.id


@pytest.mark.asyncio
async def test_session_clear(
    client: httpx.Client
):
    response = await client.post(url='/')
    assert settings.SESSION_COOKIE_NAME in response.impl.cookies
    session1 = Session.parse_cookie(response.impl.cookies[settings.SESSION_COOKIE_NAME])
    assert session1 is not None

    response = await client.delete(url='/')
    assert settings.SESSION_COOKIE_NAME in response.impl.cookies
    session2 = Session.parse_cookie(response.impl.cookies[settings.SESSION_COOKIE_NAME])
    assert session2 is not None
    assert session1.id != session2.id


class SessionEndpoint(cbra.Endpoint):

    async def get(self) -> dict[str, Any]:
        await self.session
        return self.session.data.dict()

    async def post(self) -> dict[str, Any]:
        await self.session
        self.session.set('iss', 'foo')
        self.session.set('sub', 'bar')
        return self.session.data.dict()

    async def delete(self) -> dict[str, Any]:
        await self.session
        await self.session.clear()
        return self.session.data.dict()


@pytest_asyncio.fixture # type: ignore
async def client():
    app = cbra.Application()
    app.add(SessionEndpoint)
    async with httpx.Client(base_url='https://cbra', app=app) as client:
        yield client


# Copyright (C) 2021-2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Any

import pytest
import pytest_asyncio
from headless.core import httpx

import cbra.core as cbra


class MockEndpoint(cbra.Endpoint):

    async def head(self) -> dict[str, Any]:
        return {'method': self.request.method}

    async def get(self) -> dict[str, Any]:
        return {'method': self.request.method}

    async def post(self) -> dict[str, Any]:
        return {'method': self.request.method}

    async def put(self) -> dict[str, Any]:
        return {'method': self.request.method}

    async def patch(self) -> dict[str, Any]:
        return {'method': self.request.method}

    async def delete(self) -> dict[str, Any]:
        return {'method': self.request.method}


class AnnotationEndpoint(cbra.Endpoint):

    async def post(self, data: dict[str, Any]) -> dict[str, Any]:
        return {
            'method': self.request.method,
            'headers': {str.lower(k): v for k, v in self.request.headers.items()},
            'data': data
        }


@pytest_asyncio.fixture # type: ignore
async def client():
    app = cbra.Application()
    app.add(MockEndpoint, path='/')
    app.add(AnnotationEndpoint, path='/annotations')

    async with httpx.Client(base_url='https://cbra', app=app) as client:
        yield client


@pytest.mark.parametrize("method", [
    "HEAD",
    "GET",
    "POST",
    "PUT",
    "PATCH",
    "DELETE"
])
@pytest.mark.asyncio
async def test_basic_request_methods(
    client: httpx.Client,
    method: str
):
    response = await client.request(
        method=method,
        url='/',
    )
    assert response.status_code == 200
    if method != 'HEAD':
        assert response.headers.get('Content-Type') == 'application/json'
        dto = await response.json()
        assert dto.get('method') == method


@pytest.mark.parametrize("method", [
    "HEAD",
    "GET",
    "POST",
    "PUT",
    "PATCH",
    "DELETE"
])
@pytest.mark.asyncio
async def test_options(
    client: httpx.Client,
    method: str
):
    response = await client.options(url='/')
    assert response.allows(method)


# Test that a function can use f(body_json: dict[str, Any])
@pytest.mark.asyncio
async def test_dict_from_json_without_content_type(
    client: httpx.Client
):
    response = await client.post(url='/annotations', content='{"foo": 1}')
    assert response.status_code == 200
    result = await response.json()
    assert 'data' in result
    assert 'headers' in result
    assert 'content-type' not in result['headers']
    assert 'foo' in result['data']


@pytest.mark.asyncio
async def test_dict_from_json_with_content_type(
    client: httpx.Client
):
    response = await client.post(
        url='/annotations',
        content='{"foo": 1}',
        headers={'Content-Type': "application/json"}
    )
    assert response.status_code == 200
    result = await response.json()
    assert 'data' in result
    assert 'headers' in result
    assert 'content-type' in result['headers']
    assert 'foo' in result['data']


@pytest.mark.asyncio
async def test_dict_from_invalid_json(
    client: httpx.Client
):
    response = await client.post(url='/annotations', content='{"foo"}: 1}')
    assert response.status_code == 422


# Copyright (C) 2022 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from types import NotImplementedType
from typing import Any
from typing import TypeVar

import pydantic
import pydantic.main

from .modelidentity import ModelIdentity
from .modelautoassignedidentity import ModelAutoAssignedIdentity


T = TypeVar('T', bound='BaseModel')


class BaseModelMetaclass(pydantic.main.ModelMetaclass):
    """Adds some additional features to a :mod:`pydantic` model:
    
    * Fields accept the ``primary_key`` argument, indicating that the
      field is part of the identity of a model.
    * Fields accept the ``auto_assign`` argument, indicating that the
      field may be auto assigned by, for example, a storage backend.
      This is only supported for :class:`int` fields.
    """
    __abstract__: bool
    __module__: str = 'cbra.types'

    def __new__(
        cls,
        name: str,
        bases: tuple[type, ...],
        namespace: dict[str, Any],
        **params: Any
    ) -> 'BaseModel':
        for b in bases:
            if isinstance(b, BaseModelMetaclass) and not b.__abstract__:
                raise NotImplementedError("Subclassing is not implemented.")

        # Collect the surrogate key, if defined. Multiple surrogate
        # keys are an error.
        if '__surrogate__' in namespace:
            raise TypeError(f'{name}.__surrogate__ must not be declared.')
        namespace.setdefault('__surrogate__', NotImplemented)
        for attname, value in namespace.items():
            if not isinstance(value, pydantic.fields.FieldInfo)\
            or not value.extra.get('auto_assign'):
                continue
            if value.extra.get('primary_key'):
                raise TypeError(
                    f'{name}.{attname} can not be part of both the primary key '
                    'and the surrogate key.'
                )
            if namespace.get('__surrogate__') != NotImplemented:
                raise TypeError(f'{name} can not have multiple auto_assigned keys.')
            namespace['__surrogate__'] = ModelAutoAssignedIdentity(attname, value)

        new_class: 'BaseModel' = super().__new__( # type: ignore
            cls, name, bases, namespace, **params
        )
        fields: dict[str, pydantic.fields.FieldInfo] = {}
        fields.update({
            attname: field.field_info
            for attname, field in new_class.__fields__.items()
        })

        # Collect all fields that are marked as keys and create
        # ModelIdentity instances so we can use them later to
        # build queries and URLs. For now we only support one
        # key per model but natural keys also need to be supported
        # at one point.
        key_fields: list[pydantic.fields.FieldInfo] = []
        for name, field in fields.items():
            if not field.extra.get('primary_key'):
                continue
            field.extra['name'] = name
            key_fields.append(field)
        if key_fields:
            new_class.__key__ = ModelIdentity(key_fields)

        return new_class # type: ignore


class BaseModel(pydantic.BaseModel, metaclass=BaseModelMetaclass):
    __abstract__: bool = True
    __key__: ModelIdentity | NotImplementedType
    __surrogate__: ModelAutoAssignedIdentity | NotImplementedType
    __module__: str = 'cbra.types'

    def __eq__(self: T, other: T) -> bool:
        # TODO: Compare subclasses to parents if the key is the
        # same?
        return all([
            type(self) == type(other),
            self.__key__ == other.__key__ # type: ignore
        ])

    def __hash__(self) -> int: # type: ignore
        # TODO: This is implemented to support set operations,
        # but the object is mutable and thus can lead to a set
        # with multiple objects of the same identity.
        return hash(self.__key__)


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import logging
from typing import Any

import aorta
import fastapi

from cbra.types import IDependant
from .ioc import instance
from .messagerunner import MessageRunner


class MessagePublisher(aorta.MessagePublisher, IDependant):
    __module__: str = 'cbra.core'
    logger: logging.Logger = logging.getLogger('aorta')
    request: fastapi.Request
    runner: MessageRunner

    def __init__(
        self,
        request: fastapi.Request,
        transport: aorta.types.ITransport = instance('MessageTransport')
    ):
        super().__init__(transport=transport)
        self.request = request

    async def send(
        self,
        messages: list[aorta.types.Envelope[Any]],
        is_retry: bool = False
    ) -> None:
        for message in messages:
            self.logger.debug(
                "Publishing message %s/%s (id: %s)",
                message.api_version, message.kind, message.metadata.uid
            )
        return await super().send(messages, is_retry)


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import get_args
from typing import Callable

import fastapi
from headless.ext.picqer import Client
from headless.ext.picqer import DefaultClient
from headless.ext.picqer.v1 import Webhook

import cbra.core
from cbra import types
from cbra.core.conf import settings

from .webhooksignature import WebhookSignature
from .webhookresponse import WebhookResponse
from .webhooksecret import WebhookSecret
from .v1 import PicqerWebhookMessage


class PicqerWebhookEndpoint(cbra.core.Endpoint):
    __module__: str = 'cbra.ext.picqer'
    client: Client = fastapi.Depends(DefaultClient)
    tags: list[str] = ['Integrations']
    name: str = 'picqer.webhooks'
    summary: str = 'Picqer Webhooks'
    with_options: bool = False
    signature: WebhookSignature = WebhookSignature.depends()

    @classmethod
    async def install(
        cls,
        client: Client,
        callback_url: str,
        generate_name: Callable[[str], str],
        secret: str | None = None,
        skip_missing: bool = True
    ):
        """Install webhooks created by this server."""
        # Inspect the types to get the supported Picqer event types.
        events: set[str] = set()
        for model in get_args(PicqerWebhookMessage):
            for event in get_args(model.__fields__['event'].type_):
                fn = f"on_{str.replace(event, '.', '_')}"
                if not hasattr(cls, fn) and skip_missing:
                    cls.logger.debug(
                        'Event %s is not implemented by %s',
                        event, cls.__name__
                    )
                    continue
                events.add(event)

        # Fetch all webhooks and determine determine which ones we
        # need to create.
        async for hook in client.listall(Webhook):
            if not hook.active:
                continue
            if hook.name != generate_name(hook.event):
                continue
            events.remove(hook.event)
        
        # The remaining hooks need to be created.
        for event in sorted(events):
            cls.logger.warning("Creating webhook for event %s", event)
            params: dict[str, str] = {
                'name': generate_name(event),
                'event': event,
                'address': callback_url
            }
            if secret:
                params['secret'] = secret
            await client.post(
                url=Webhook.get_create_url(),
                json=params
            )

    async def post(self, dto: PicqerWebhookMessage) -> WebhookResponse:
        if not await self.signature.verify(await self.get_webhook_secret()):
            raise types.Forbidden
        attname = f"on_{str.replace(dto.event, '.', '_')}"
        if not hasattr(self, attname):
            self.logger.critical(
                'Picqer event %s is not implemented by this endpoint.',
                dto.event
            )
            return self.reject(reason=f"This server does not implement {dto.event}")
        func = getattr(self, attname)
        try:
            return await func(dto) or WebhookResponse(
                accepted=True,
                success=True
            )
        except Exception:
            self.logger.exception("Caught fatal exception while handling webhook message")
            return WebhookResponse(
                accepted=True,
                success=False
            )

    async def get_webhook_secret(self) -> types.Verifier:
        return WebhookSecret(settings.PICQER_WEBHOOK_SECRET) # type: ignore

    def reject(self, reason: str) -> WebhookResponse:
        return WebhookResponse(accepted=False, success=False, reason=reason)


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from cbra.core.conf import settings
from cbra.ext import cache
from cbra.types import ICache
from cbra.types import ISubject
from cbra.types import ISubjectResolver
from .resource.types import UserInfoSubject
from .params import RequestAccessToken


class ResourceServerSubjectResolver(ISubjectResolver):
    __module__: str = 'cbra.ext.oauth2'
    storage: ICache

    def __init__(
        self,
        storage: ICache = cache.cache(settings.OAUTH2_CACHE)
    ):
        self.storage = storage

    def cache_key(self, principal: RequestAccessToken):
        return f'oauth2:access-token:userinfo:{principal.token_id}'

    async def resolve( # type: ignore
        self,
        principal: RequestAccessToken
    ) -> ISubject:
        assert principal.token_id is not None
        userinfo = await self.storage.get(self.cache_key(principal))
        if userinfo is None:
            userinfo = await principal.userinfo()
            await self.storage.set(self.cache_key(principal), userinfo.dict())
        return UserInfoSubject(userinfo)


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import fastapi

import cbra.core as cbra
from ..types import ExtAuthorizationRequestState
from ..types import FatalAuthorizationException
from ..types import IAuthorizationServerStorage
from ..types import IExternalAuthorizationState


__all__: list[str] = [
    'LocalAuthorizationRequestState'
]


async def get(
    storage: IAuthorizationServerStorage = cbra.instance(
        name='_AuthorizationServerStorage'
    ),
    state: ExtAuthorizationRequestState | None = fastapi.Query(
        default=None,
        title='State',
        description=(
            'An opaque value used by the client to maintain state '
            'between the request and callback. The authorization '
            'server includes this value when redirecting the '
            'user-agent back to the client. If supplied, this '
            'parameter **must** equal the `state` parameter used when '
            'creating the authorization request. The `state` parameter '
            'is ignored when using JARM because it is included in the '
            'JSON Web Token supplied using the `jwt` parameter per '
            'chosen response mode.'
        )
    ),
):
    if state is None:
        raise FatalAuthorizationException("The state parameter is missing or invalid.")
    current = await storage.fetch(state)
    if current is None:
        raise FatalAuthorizationException("The state parameter is missing or invalid.")
    return current


LocalAuthorizationRequestState: IExternalAuthorizationState = fastapi.Depends(get)


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from .accesscodeextension import AccessCodeExtension
from .baseemailsender import BaseEmailSender
from .buttonextension import ButtonExtension
from .paragraphextension import ParagraphExtension
from .sectionextension import SectionExtension


__all__: list[str] = [
    'AccessCodeExtension',
    'BaseEmailSender',
    'ButtonExtension',
    'ParagraphExtension',
    'SectionExtension'
]


# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
# import os
# import sys
# sys.path.insert(0, os.path.abspath('.'))


# -- Project information -----------------------------------------------------
import os
import sys
from os.path import abspath
from os.path import dirname
from os.path import join

project = 'CBRA'
copyright = '2021-2023, Cochise Ruhulessin'
author = 'Cochise Ruhulessin'

# The full version, including alpha/beta/rc tags
release = os.getenv('SEMVER_RELEASE') or '0.0.1'

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
sys.path.append(abspath(join(dirname(__file__), "_ext")))

# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    'djangodocs',
    'sphinx.ext.autodoc',
    'sphinx.ext.intersphinx',
    'sphinx.ext.napoleon',
    'sphinx_copybutton'
]

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = []


# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = 'sphinx_material'

html_sidebars = {
    "**": ["logo-text.html", "globaltoc.html", "localtoc.html", "searchbox.html"]
}

html_static_path = ['_static']

# These paths are either relative to html_static_path
# or fully qualified paths (eg. https://...)
html_css_files = [
    'css/style.css',
]

html_theme_options = {
    'nav_title': 'CBRA',
    'google_analytics_account': 'UA-XXXXX',
    'base_url': 'https://gitlab.com/unimatrixone/libraries/python-unimatrix/cbra',
    'color_primary': 'blue',
    'color_accent': 'light-blue',
    'repo_url': 'https://github.com/cochiseruhulessin/python-cbra',
    'repo_name': 'CBRA',
    'globaltoc_depth': 2,
    'globaltoc_collapse': True,
    'globaltoc_includehidden': True,
}

add_module_names = False


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Any
from typing import Protocol

from headless.ext.oauth2.types import GrantType

from cbra.core.iam.models import Subject
from .iauthorizationrequest import IAuthorizationRequest
from .iresourceowner import IResourceOwner
from .refreshtokenpolicytype import RefreshTokenPolicyType
from .rfc9068accesstoken import RFC9068AccessToken
from .requestedscope import RequestedScope
from .signableoidctoken import SignableOIDCToken


class ITokenBuilder(Protocol):
    __module__: str = 'cbra.ext.oauth2.types'

    def rfc9068(
        self,
        sub: int,
        scope: list[RequestedScope],
        auth_time: int,
        audience: str | None = None,
    ) -> tuple[RFC9068AccessToken, int]:
        ...

    async def refresh_token(
        self,
        grant_type: GrantType,
        client_id: str,
        sector_identifier: str,
        sub: int,
        ppid: int,
        claims: dict[str, Any],
        scope: list[RequestedScope],
        renew: RefreshTokenPolicyType,
        ttl: int,
        auth_time: int,
        resources: list[str] | None = None
    ) -> str:
        ...

    def id_token(
        self,
        subject: Subject,
        ppid: int,
        nonce: str,
        scope: list[RequestedScope],
        access_token: str | None,
        auth_time: int | None = None,
        authorization_code: str | None = None,
        request: IAuthorizationRequest | None = None,
        owner: IResourceOwner | None = None,
    ) -> SignableOIDCToken:
        ...

    async def userinfo(
        self,
        subject: Subject,
        owner: IResourceOwner,
        request: IAuthorizationRequest,
        scope: list[RequestedScope],
    ) -> dict[str, Any]:
        ...


# Copyright (C) 2020-2022 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
"""Declares :class:`GoogleTransport`."""
from typing import Any

from aorta.types import Envelope
from aorta.ext.google import PubsubTransport
import fastapi

from cbra.core.conf import settings


class GooglePubsubTransport(PubsubTransport):
    prefix: str
    service_name: str

    def __init__(
        self,
        project: str = fastapi.Depends(lambda: settings.GOOGLE_HOST_PROJECT),
        prefix: str = fastapi.Depends(lambda: settings.PUBLISHER_TOPIC_PREFIX),
        service_name: str = fastapi.Depends(lambda: settings.APP_NAME)
    ):
        self.prefix = prefix
        self.service_name = service_name
        super().__init__(
            project=project,
            topic=self.topic_factory,
            retry_topic=f'{self.prefix}.retry.{service_name}'
        )

    def topic_factory(self, envelope: Envelope[Any]) -> list[str]:
        if envelope.is_event():
            topics: list[str] = [
                f'{self.prefix}.events.{envelope.kind}'
            ]
            if not envelope.is_private_event():
                topics.append(f'{self.prefix}.events')
        elif envelope.is_command():
            topics = [f'{self.prefix}.commands.{self.service_name}']
            if envelope.metadata.audience:
                topics = [
                    f'{self.prefix}.commands.{x}' if x != 'self'\
                    else f'{self.prefix}.commands.{self.service_name}'
                    for x in envelope.metadata.audience
                    if x != 'self'
                ]
        else:
            raise NotImplementedError
        return topics


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from .accesscontrollist import AccessControlList
from .externalprincipal import ExternalPrincipal
from .principal import Principal
from .subject import Subject
from .subjectclaimset import SubjectClaimSet


__all__: list[str] = [
    'AccessControlList',
    'ExternalPrincipal',
    'Principal',
    'Subject',
    'SubjectClaimSet',
]


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import functools
from typing import Any

from google.cloud.datastore import Client

from .runner import Runner
from .types import IDatastoreCursor
from .types import IDatastoreEntity
from .types import IDatastoreKey
from .types import IDatastoreQuery


class BaseDatastoreRepository(Runner):
    __module__: str = 'cbra.ext.google'

    def __init__(
        self,
        client: Client
    ) -> None:
        self.client = client

    async def allocate(self, name: str) -> int:
        base = self.key(kind=name)
        result = await self.run_in_executor(
            functools.partial( # type: ignore
                self.client.allocate_ids, # type: ignore
                incomplete_key=base,
                num_ids=1
            )
        )
        return [x for x in result][0].id

    def key(
        self,
        kind: str,
        identifier: int | str | None = None,
        parent: IDatastoreKey | None = None
    ) -> IDatastoreKey:
        args: list[Any] = [kind]
        if identifier is not None:
            args.append(identifier)
        return self.client.key(*args, parent=parent) # type: ignore

    def entity_factory(self, key: IDatastoreKey, **attrs: Any) -> IDatastoreEntity:
        entity = self.client.entity(key) # type: ignore
        if attrs:
            entity.update(attrs) # type: ignore
        return entity # type: ignore

    def entity(
        self,
        kind: str,
        key: int | str | IDatastoreKey | None = None,
        parent: IDatastoreKey | None = None
    ) -> IDatastoreEntity:
        if isinstance(key, (int, str)):
            key = self.key(kind, key)
        return self.client.entity(key or self.key(kind, parent=parent)) # type: ignore

    def query(self, *args: Any, **kwargs: Any) -> IDatastoreQuery:
        return self.client.query(*args, **kwargs) # type: ignore

    async def delete(
        self,
        keys: IDatastoreKey | list[IDatastoreKey]
    ) -> None:
        if not isinstance(keys, list):
            keys = [keys]
        return await self.run_in_executor(
            functools.partial(self.client.delete_multi, keys) # type: ignore
        )

    async def execute(self, query: IDatastoreQuery, **kwargs: Any) -> IDatastoreCursor:
        cursor = await self.run_in_executor(functools.partial(query.fetch, **kwargs))
        return cursor

    async def get_entity_by_key(self, key: IDatastoreKey) -> IDatastoreEntity | None:
        return await self.run_in_executor(
            functools.partial(
                self.client.get, # type: ignore
                key=key
            )
        )
    
    async def first(self, query: IDatastoreQuery) -> IDatastoreEntity | None:
        """Return the first object in the query, or none."""
        for entity in await self.execute(query, limit=1):
            return entity

    async def put(self, entity: IDatastoreEntity) -> IDatastoreKey:
        await self.run_in_executor(self.client.put, entity) # type: ignore
        return entity.key # type: ignore


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Any

from headless.ext.oauth2.models import ClaimSet

from cbra.types import ISubject


class UserInfoSubject(ISubject):
    """Describes a :term:`Resource Owner` of which the claims were
    received from the OpenID Connect UserInfo endpoint.
    """
    __module__: str = 'cbra.ext.oauth2.resource.types'
    claims: ClaimSet

    def __init__(self, claims: dict[str, Any] | ClaimSet):
        if isinstance(claims, dict):
            claims = ClaimSet.parse_obj(claims)
        self.claims = claims

    def is_authenticated(self) -> bool:
        return True

    def get_display_name(self) -> str:
        raise NotImplementedError


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from headless.ext import oauth2

from ..types import IAccessToken
from ..types import IAccessTokenObtainer
from ..types import IFrontendStorage
from ..types import IManagedGrant


class AccessTokenFactory(IAccessTokenObtainer):
    __module__: str = 'cbra.ext.oauth2.client'
    client: oauth2.Client
    grant: IManagedGrant
    resource: str
    storage: IFrontendStorage

    def __init__(
        self,
        storage: IFrontendStorage,
        client: oauth2.Client,
        grant: IManagedGrant,
        resource: str
    ):
        self.client = client
        self.grant = grant
        self.resource = resource
        self.storage = storage

    async def obtain(self, scope: set[str] | None = None) -> IAccessToken:
        return await self.grant.refresh(
            storage=self.storage,
            client=self.client,
            resource=self.resource,
            scope=scope
        )


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
"""
.. _ref-guides-integrating-an-oauth2-authorization-server

==============================================
Implementing an OAuth 2.x authorization server
==============================================
"""
import copy
from typing import Any

from headless.ext import oauth2

import cbra.core as cbra
from cbra.core.conf import settings
from cbra.core.iam.types import ISubjectRepository
from cbra.core.iam.types import IUserOnboardingService
from cbra.core.iam.types import Subject
from cbra.core.iam.services import UserOnboardingService
from ..authorizationserverstorage import AuthorizationServerStorage
from ..models import ResourceOwner
from ..types import IClient
from ..types import ResourceOwnerIdentifier


class AuthorizationServerEndpoint(cbra.Endpoint):
    path: str
    tags: list[str] = ['OAuth 2.x/OpenID Connect']
    with_options: bool = False
    client: IClient
    onboard: IUserOnboardingService = cbra.instance(
        name='SubjectOnboardingService',
        missing=UserOnboardingService
    )
    storage: AuthorizationServerStorage = cbra.instance('_AuthorizationServerStorage')
    subjects: ISubjectRepository = cbra.instance('SubjectRepository')

    def delete_cookies(self, exclude: set[str] | None= None):
        """Deletes all cookies set by the authorization server."""
        exclude = exclude or set()
        for k in self.request.cookies:
            if not str.startswith(k, 'oauth2') or k in exclude:
                continue
            self.delete_cookie(k)

    async def get_client(self, client_id: str) -> oauth2.Client | None:
        """Return a preconfigured OAuth 2.x/OpenID Connect client,
        or ``None`` if the client does not exist.
        """
        # TODO: Quite ugly
        for client in settings.OAUTH2_CLIENTS:
            if client_id in {client.get('name'), client.get('client_id')}:
                instance = oauth2.Client(**copy.deepcopy(client))
                break
        else:
            instance = None
        return instance

    def get_issuer(self) -> str:
        return settings.OAUTH2_ISSUER or\
            f'{self.request.url.scheme}://{self.request.url.netloc}'

    async def get_owner(self, client_id: str) -> ResourceOwner | None:
        await self.session
        oid = ResourceOwnerIdentifier(client_id=client_id, sub=int(self.session.uid))
        return await self.storage.get(ResourceOwner, oid)

    async def get_subject(self) -> Subject:
        if not self.is_authenticated():
            raise ValueError("The request is not authenticated")
        assert self.session.uid is not None
        subject = await self.storage.get_subject(self.session.uid)
        assert subject is not None
        return subject
    
    async def persist(self, obj: Any) -> None:
        await self.storage.persist(obj)



# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Protocol

from .oidcprovider import OIDCProvider
from .pairwiseidentifier import PairwiseIdentifier
from .redirecturi import RedirectURI


class IClient(Protocol):
    __module__: str = 'cbra.ext.oauth2.types'
    client_id: str
    sector_identifier: str

    def can_redirect(self, uri: RedirectURI) -> bool: ...
    def get_pairwise_identifier(self, sub: int) -> PairwiseIdentifier: ...
    def get_provider(self) -> OIDCProvider: ...
    def get_redirect_uri(self, uri: RedirectURI | None) -> RedirectURI: ...
    def requires_downstream(self) -> bool: ...


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Any

from cbra.types import IQueryResult
from .resourcemodel import ResourceModel


class Collection:
    __module__: str = 'cbra.core'
    cursor_param_name: str = '_c'
    model: type[ResourceModel]

    async def list(self) -> Any:
        result = await self.filter(None)
        resource =  self.model.__list_model__.parse_obj({
            'apiVersion': 'v1',
            'kind': f'{self.model.__name__}List',
            'metadata': {
                'nextUrl': self.get_next_url(result.token)
            },
            'items': []
        })
        resource.items = [self.adapt(x) for x in result.items] # type: ignore
        return resource

    async def filter(self, params: Any) -> IQueryResult[Any]:
        raise NotImplementedError
    
    def get_next_url(self, token: str | None) -> str | None:
        if not token:
            return None
        return self.reverse('list', params={self.cursor_param_name: token}) # type: ignore


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import fastapi

from cbra.types import Abortable


class BearerTokenException(Abortable):
    error: str
    error_description: str
    scheme: str
    status_code: int = 401

    def __init__(self, scheme: str, error:str, error_description: str, status_code: int = 401) -> None:
        self.error = error
        self.error_description = error_description
        self.scheme = scheme
        self.status_code = status_code

    async def as_response(self) -> fastapi.Response:
        response = fastapi.Response(status_code=self.status_code)
        value = self.scheme
        if self.error:
            value += f', error="{self.error}"'
        if self.error_description:
            value += f', error_description="{self.error_description}"'
        response.headers['WWW-Authenticate'] = value
        return response



# Copyright (C) 2021-2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import collections.abc
import json
import logging
from typing import Any


class JSONFormatter(logging.Formatter):
    __module__: str = 'cbra.ext.google.logging'

    def format(self, record: logging.LogRecord) -> str:
        super().format(record)
        params: dict[str, Any] = {
            'message': record.msg
        }
        if isinstance(record.msg, collections.abc.Mapping):
            params.update(record.msg)
        else:
            params['message'] = super().formatMessage(record)
        if record.exc_info:
            assert record.exc_info[0] is not None
            params['exception'] = {
                'type': f'{record.exc_info[0].__module__}.{record.exc_info[0].__name__}',
                'stack': self.formatException(record.exc_info)
            }
        return json.dumps(params)


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import secrets
from datetime import datetime
from datetime import timezone
from typing import Any

import fastapi
import pydantic
from canonical import EmailAddress
from headless.ext.oauth2.models import OIDCToken
from headless.ext.oauth2.types import ResponseType

from cbra.types import SessionClaims
from ..types import AccessType
from ..types import AuthorizationRequestIdentifier
from ..types import ClientInfo
from ..types import RedirectURI
from ..types import RequestedScope
from .authorizationcode import AuthorizationCode
from .authorizationrequestclient import AuthorizationRequestClient


class AuthorizationRequestParameters(pydantic.BaseModel):
    access_type: AccessType = AccessType.online
    client_id: str
    code: AuthorizationCode | None = None
    id: AuthorizationRequestIdentifier | None = None
    response_type: ResponseType
    redirect_uri: RedirectURI
    scope: list[RequestedScope] = []
    state: str | None = None
    nonce: str | None = None
    resource: list[str] = []

    # Internal properties
    auth_time: int | None = None
    client_info: ClientInfo
    consent: list[RequestedScope] = []
    email: EmailAddress | None = None
    email_verified: bool | None = None
    id_token: dict[str, Any] = {}
    session_id: str | None = None
    downstream: OIDCToken | None = None
    owner: int | None = None
    requested: datetime = pydantic.Field(
        default_factory=lambda: datetime.now(timezone.utc)
    )
    remote_host: str
    session: SessionClaims | None = None

    @pydantic.root_validator(pre=True)
    def preprocess(
        cls,
        values: dict[str, Any]
    ) -> dict[str, Any]:
        client: AuthorizationRequestClient = values.pop('client', None)
        redirect_uri: RedirectURI | None = values.get('redirect_uri')
        scope: str = values.get('scope') or ''
        if redirect_uri is None and client is not None:
            values['redirect_uri'] = client.get_redirect_uri(redirect_uri)
        if scope and isinstance(scope, str):
            values['scope'] = set(filter(bool, str.split(scope, ' ')))
        return values

    async def load(
        self,
        client: Any,
        storage: Any,
        session_id: str
    ):
        if self.id is None:
            self.id = AuthorizationRequestIdentifier(secrets.token_urlsafe(48))
        if self.issues_authorization_code():
            self.code = AuthorizationCode()
        return self

    def as_response(
        self,
        client: AuthorizationRequestClient,
        issuer: str
    ) -> fastapi.Response:
        assert self.code is not None
        if self.response_type != ResponseType.code:
            raise NotImplementedError
        url = self.redirect_uri.redirect(
            code=self.code.value,
            state=self.state,
            iss=issuer
        )
        assert isinstance(url, str)
        return fastapi.responses.RedirectResponse(
            status_code=303,
            url=url,
        )

    def assign(self, owner: int) -> None:
        """Set the owner for the Authorization Request. This will be used
        later when exchanging the authorization code.
        """
        self.owner = owner

    def issues_authorization_code(self) -> bool:
        return self.response_type in {
            ResponseType.code,
            ResponseType.code_id_token,
            ResponseType.code_id_token_token,
            ResponseType.code_token
        }


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from datetime import datetime
from datetime import timezone
from typing import Any

from headless.ext.oauth2.types import GrantType

import cbra.core as cbra
from cbra.core.iam.models import Subject
from cbra.types import IDependant
from ..authorizationserverstorage import AuthorizationServerStorage
from ..models import Client
from ..models import RefreshToken
from ..types import IAuthorizationRequest
from ..types import IResourceOwner
from ..types import RefreshTokenPolicyType
from ..types import RequestedScope
from ..types import RFC9068AccessToken
from ..types import SignableOIDCToken
from .currentissuer import CurrentIssuer
from .requestingclient import RequestingClient


class TokenBuilder(IDependant):
    __module__: str = 'cbra.ext.oauth2.params'
    access_token_ttl: int = 3600
    id_token_ttl: int = 600
    client: Client
    issuer: str
    now: int
    storage: AuthorizationServerStorage

    def __init__(
        self,
        storage: AuthorizationServerStorage = cbra.instance('_AuthorizationServerStorage'),
        issuer: str = CurrentIssuer,
        client: Client = RequestingClient
    ) -> None:
        self.client = client
        self.issuer = issuer
        self.now = int(datetime.now(timezone.utc).timestamp())
        self.storage = storage

    def id_token(
        self,
        subject: Subject,
        ppid: int,
        nonce: str,
        scope: list[RequestedScope],
        access_token: str | None,
        auth_time: int | None = None,
        authorization_code: str | None = None,
        request: IAuthorizationRequest | None = None,
        owner: IResourceOwner | None = None,
    ) -> SignableOIDCToken:
        claims: dict[str, Any] = {
            'aud': self.client.client_id,
            'auth_time': auth_time,
            'exp': self.now + self.id_token_ttl,
            'azp': self.client.client_id,
            'iat': self.now,
            'iss': self.issuer,
            'nonce': nonce,
            'sub': '0'
        }
        for claim in scope:
            claim.apply(
                subject=subject,
                claims=claims,
                owner=owner, # type: ignore
                request=request,
            )
        return SignableOIDCToken.parse_obj({
            'access_token': access_token,
            'authorization_code': authorization_code,
            'claims': claims
        })

    def rfc9068(
        self,
        sub: int,
        scope: list[RequestedScope],
        auth_time: int,
        audience: str | None = None,
    ) -> tuple[RFC9068AccessToken, int]:
        ttl = self.client.access_token_ttl or self.access_token_ttl
        token = RFC9068AccessToken.new(
            client_id=self.client.client_id,
            iss=self.issuer,
            aud=audience or self.issuer,
            sub=str(sub),
            scope=scope,
            auth_time=auth_time,
            now=self.now,
            ttl=ttl
        )
        return token, ttl

    async def refresh_token(
        self,
        grant_type: GrantType,
        client_id: str,
        sector_identifier: str,
        sub: int,
        ppid: int,
        claims: dict[str, Any],
        scope: list[RequestedScope],
        renew: RefreshTokenPolicyType,
        ttl: int,
        auth_time: int,
        resources: list[str] | None = None
    ) -> str:
        instance = RefreshToken.parse_obj({
            'grant_type': grant_type,
            'auth_time': auth_time,
            'claims': claims,
            'client_id': client_id,
            'sector_identifier': sector_identifier,
            'sub': sub,
            'ppid': ppid,
            'scope': list(sorted([x.name for x in scope])),
            'renew': renew,
            'resources': resources or [],
            'ttl': ttl
        })
        await self.storage.persist(instance)
        return instance.token

    async def userinfo(
        self,
        subject: Subject,
        owner: IResourceOwner,
        request: IAuthorizationRequest,
        scope: list[RequestedScope],
    ) -> dict[str, Any]:
        claims: dict[str, Any] = {
            'auth_time': request.auth_time,
            'sub': owner.ppid.value,
            'sct': owner.ppid.sector
        }
        for claim in scope:
            claim.apply(
                subject=subject,
                claims=claims,
                owner=owner, # type: ignore
                request=request,
            )
        return claims


# Copyright (C) 2022 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import functools
import warnings
from typing import Any

from canonical import EmailAddress
from canonical import StringType


class PolicyPrincipal(StringType):
    """Identifies a member (principal) in an access control list or
    role membership. A :class:`PolicyPrincipal` may be one of the following
    types:

    - `domain` - a specific domain. Domain membership is established by
      proving control over an email address for the domain.
    - `user` - a specific user, identified by email address.
    - `sub` - a specific subject, identified by subject identifier.
    """
    __module__: str = 'cbra.types'
    principal_types: set[str] = {
        'domain',
        'user',
        'sub',
    }

    @classmethod
    def validate(cls, v: str) -> str:
        v = super().validate(v)
        t, _ = str.split(v, ':', 1)
        if t not in cls.principal_types:
            raise ValueError(f'invalid principal type: {t}')
        return cls(v)
    
    @functools.singledispatchmethod
    def match(self, obj: Any) -> bool:
        warnings.warn(
            f"Unknown princpal for access policy: {type(obj).__name__}"
        )
        return False
    
    @match.register
    def match_email(self, email: EmailAddress) -> bool:
        kind, value = str.split(self, ':', 1)
        if kind not in {'domain', 'user', 'sub'}:
            return False
        is_matching = False
        if kind == 'domain':
            is_matching = value == str(email.domain)
        elif kind == 'user':
            is_matching = value == str(email)
        return is_matching



# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import fastapi

import cbra.core as cbra
from ..types import FatalAuthorizationException
from ..types import IAuthorizationRequest
from ..types import IAuthorizationServerStorage
from ..types import IExternalAuthorizationState
from .localauthorizationrequeststate import LocalAuthorizationRequestState


__all__: list[str] = [
    'LocalStorageAuthorizationRequest'
]


async def get(
    storage: IAuthorizationServerStorage = cbra.instance(
        name='_AuthorizationServerStorage'
    ),
    state: IExternalAuthorizationState = LocalAuthorizationRequestState
):
    authnrequest = await storage.fetch(state.request_id)
    if authnrequest is None:
        raise FatalAuthorizationException(
            f"The authorization request (id: {state.request_id}) does not "
            "exist, is expired, or otherwise not processable."
        )
    return authnrequest


LocalStorageAuthorizationRequest: IAuthorizationRequest = fastapi.Depends(get)


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Generic
from typing import TypeVar


T = TypeVar('T')


class DatastoreQueryResult(Generic[T]):
    __module__: str = 'cbra.ext.google'
    token: str | None
    items: list[T]

    def __init__(self, token: str | None, entities: list[T]) -> None:
        self.token = token or None
        self.items = entities


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import base64
import hashlib

import pydantic


class SubjectIdentifier(pydantic.BaseModel):
    iss: str
    sub: str

    @property
    def sha256(self) -> str:
        h = hashlib.sha3_256()
        h.update(str.encode(type(self).__name__))
        h.update(str.encode(self.iss))
        h.update(str.encode(self.sub))
        return bytes.decode(base64.urlsafe_b64encode(h.digest()))


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Any
from typing import Protocol
from typing import TypeVar

from .ipolymorphiccursor import IPolymorphicCursor
from .ipolymorphicquery import IPolymorphicQuery
from .persistedmodel import PersistedModel


T = TypeVar('T', bound=PersistedModel)


class IPolymorphicRepository(Protocol):
    __module__: str = 'cbra.types'

    async def auto_increment(self, obj: type[PersistedModel] | PersistedModel) -> int:
        """Return an auto incrementing integer to be used as a technical
        primary key.
        """
        ...

    async def list(
        self,
        cls: type[T],
        ordering: list[str] | None = None,
        limit: int = 100,
        token: str | None = None
    ) -> IPolymorphicCursor[T]:
        ...
    async def query(self, cls: type[T]) -> IPolymorphicQuery[T]: ...
    async def get(self, cls: type[T], pk: Any) -> None | T: ...
    async def persist(self, obj: T) -> T: ...


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Any

import pydantic

from ..types import RFC9068AccessToken
from ..types import IAuthorizationServerStorage
from ..types import IssuedAccessToken


class BaseTokenRequest(pydantic.BaseModel):

    async def register_issued_access_token(
        self,
        sub: int,
        token: RFC9068AccessToken,
        signed: str,
        scope: list[str],
        storage: IAuthorizationServerStorage,
        claims: dict[str, Any]
    ) -> None:
        """Registers the access token as being issued, so that it
        can be later revoked or introspected.
        """
        issued = IssuedAccessToken.parse_rfc9068(
            token,
            claims=claims,
            scope=scope,
            signed_token=signed,
            sub=sub
        )
        await storage.persist(issued)


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import uvicorn
import cbra.core as cbra


URL = "https://eop5zsb1qet1zjc.m.pipedream.net"

DEFAULT_API_VERSION: str = '2023-01'

from cbra.ext.shopify import ShopifyWebhookEndpoint


app: cbra.Application = cbra.Application()
app.add(ShopifyWebhookEndpoint)
if __name__ == '__main__':
    uvicorn.run('__main__:app', reload=True) # type: ignore


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import ipaddress

from cbra.types import IAuthorizationContext
from cbra.types import ISubject


class AuthenticatedContext(IAuthorizationContext):
    __module__: str = 'cbra.core.iam'

    def __init__(
        self,
        *,
        remote_host: ipaddress.IPv4Address | str | None,
        subject: ISubject
    ):
        if isinstance(remote_host, str):
            remote_host = ipaddress.IPv4Address(remote_host)
        self._host = remote_host
        self._subject = subject

    def get_remote_host(self) -> ipaddress.IPv4Address | None:
        return self._host

    def get_subject(self) -> ISubject:
        return self._subject


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Any
from typing import Protocol

from headless.ext import oauth2
from headless.types import IClient

from .iaccesstoken import IAccessToken
from .ifrontendstorage import IFrontendStorage


class IManagedGrant(Protocol):
    __module__: str = 'cbra.ext.oauth2.types'

    def has_access_token(self, resource: str) -> bool: ...
    async def get_resource_client(
        self,
        storage: IFrontendStorage,
        client: oauth2.Client,
        resource: str,
        scope: set[str] | None = None
    ) -> IClient[Any, Any]:
        ...

    async def refresh(
        self,
        storage: IFrontendStorage,
        client: oauth2.Client,
        resource: str | None = None,
        scope: set[str] | None = None
    ) -> IAccessToken:
        ...